import { getTypeSnapshot } from "../utils/snapshot.js";
import { parseTwoslashAssertion } from "./parseTwoslashAssertion.js";
function parseAssertions(sourceFile) {
  const errorLines = /* @__PURE__ */ new Set();
  const typeAssertions = /* @__PURE__ */ new Map();
  const duplicates = [];
  const syntaxErrors = [];
  const twoSlashAssertions = [];
  const { text } = sourceFile;
  const commentRegexp = /\/\/(.*)/g;
  const lineStarts = sourceFile.getLineStarts();
  let curLine = 0;
  while (true) {
    const commentMatch = commentRegexp.exec(text);
    if (commentMatch === null) {
      break;
    }
    const comment = commentMatch[1];
    const matchExpect = /^ ?\$Expect(TypeSnapshot|Type|Error)( (.*))?$/.exec(
      comment
    );
    const commentIndex = commentMatch.index;
    const line = getLine(commentIndex);
    if (matchExpect) {
      const directive = matchExpect[1];
      const payload = matchExpect[3];
      switch (directive) {
        case "TypeSnapshot":
          const snapshotName = payload;
          if (snapshotName) {
            if (typeAssertions.delete(line)) {
              duplicates.push(line);
            } else {
              typeAssertions.set(line, {
                assertionType: "snapshot",
                expected: getTypeSnapshot(sourceFile.fileName, snapshotName),
                snapshotName
              });
            }
          } else {
            syntaxErrors.push({
              line,
              type: "MissingSnapshotName"
            });
          }
          break;
        case "Error":
          if (errorLines.has(line)) {
            duplicates.push(line);
          }
          errorLines.add(line);
          break;
        case "Type": {
          const expected = payload;
          if (expected) {
            if (typeAssertions.delete(line)) {
              duplicates.push(line);
            } else {
              typeAssertions.set(line, { assertionType: "manual", expected });
            }
          } else {
            syntaxErrors.push({
              line,
              type: "MissingExpectType"
            });
          }
          break;
        }
      }
    } else {
      const assertion = parseTwoslashAssertion(
        comment,
        commentIndex,
        line,
        text,
        lineStarts
      );
      if (assertion) {
        if ("type" in assertion) {
          syntaxErrors.push(assertion);
        } else {
          twoSlashAssertions.push(assertion);
        }
      }
    }
  }
  return {
    duplicates,
    errorLines,
    syntaxErrors,
    twoSlashAssertions,
    typeAssertions
  };
  function getLine(pos) {
    while (lineStarts[curLine + 1] <= pos) {
      curLine++;
    }
    return isFirstOnLine(text, lineStarts[curLine], pos) ? curLine + 1 : curLine;
  }
}
function isFirstOnLine(text, lineStart, pos) {
  for (let i = lineStart; i < pos; i++) {
    if (/\S/.test(text[i])) {
      return false;
    }
  }
  return true;
}
export {
  parseAssertions
};
//# sourceMappingURL=parseAssertions.mjs.map